# Copyright 2024 Yuhui. All rights reserved.
#
# Licensed under the GNU General Public License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SingStat custom types for client methods' responses."""

from datetime import date

from typing import Any, NotRequired, TypeAlias
try:
    from typing import TypedDict
except ImportError:
    TypedDict = dict

Url: TypeAlias = str
"""URL of link."""

# Resource ID

class _ResourceIdDataRecordDict(TypedDict):
    """Type definition for _ResourceIdDataDict

    All fields are required. But they are typed as `NotRequired` because
    when the endpoint returns empty data, then this dict has no keys at all.
    """
    id: NotRequired[str]
    """ID"""
    tableType: NotRequired[str]
    """Table type"""
    title: NotRequired[str]
    """Title"""

class _ResourceIdDataDict(TypedDict):
    """Type definition for ResourceIdDict"""
    generatedBy: str
    """Generated by"""
    dateGenerated: date
    """Data generated"""
    total: int
    """Total"""
    records: list[_ResourceIdDataRecordDict]
    """Records"""

class ResourceIdDict(TypedDict):
    """Type definition for resource_id()"""
    Data: _ResourceIdDataDict
    """Data"""
    DataCount: int
    """Data count"""
    StatusCode: int
    """Status code"""
    Message: str
    """Message"""

# Metadata

class _MetadataDataRecordsRowDict(TypedDict):
    """Type definition for _MetadataDataRecordsDict"""
    seriesNo: str
    """Series number"""
    rowText: str
    """Row text"""
    uoM: str
    """Unit of measurement"""
    footnote: str
    """Footnote"""

class _MetadataDataRecordsDict(TypedDict):
    """Type definition for _MetadataDataDict

    All fields are required. But they are typed as `NotRequired` because
    when the endpoint returns empty data, then this dict has no keys at all.
    """
    id: NotRequired[str]
    """ID"""
    title: NotRequired[str]
    """Title"""
    frequency: NotRequired[str]
    """Frequency"""
    dataSource: NotRequired[str]
    """Data source"""
    footnote: NotRequired[str]
    """Footnote"""
    dataLastUpdated: NotRequired[date]
    """Data last updated"""
    startPeriod: NotRequired[str]
    """Start period"""
    endPeriod: NotRequired[str]
    """End period"""
    total: NotRequired[int]
    """Total"""
    row: NotRequired[list[_MetadataDataRecordsRowDict]]
    """Row"""

class _MetadataDataDict(TypedDict):
    """Type definition for MetadataDict"""
    generatedBy: str
    """Generated by"""
    dateGenerated: date
    """Date generated"""
    records: _MetadataDataRecordsDict
    """Records"""

class MetadataDict(TypedDict):
    """Type definition for metadata()"""
    Data: _MetadataDataDict
    """Data"""
    DataCount: int
    """Data count"""
    StatusCode: int
    """Status code"""
    Message: str
    """Message"""

# Tabledata

class _TabledataDataRowColumnDict(TypedDict):
    """Type definition for _TabledataDataRowColumnColumnDict, \
        _TabledataDataTimeseriesRowDict, and \
        _TabledataDataCrossSectionalMultiDimensionalCubeRowDict
    """
    key: str
    """Key"""
    value: Any
    """Value"""

class _TabledataDataRowColumnColumnDict(TypedDict):
    """Type definition for \
        _TabledataDataCrossSectionalMultiDimensionalCubeRowDict
    """
    key: str
    """Key"""
    columns: NotRequired[list[_TabledataDataRowColumnDict]]
    """Columns"""

class _TabledataDataTimeseriesRowDict(TypedDict):
    """Type definition for _TabledataDataTimeseriesDict"""
    seriesNo: str
    """Series number"""
    rowText: str
    """Row text"""
    uoM: str
    """Unit of measurement"""
    footnote: str
    """Footnote"""
    columns: NotRequired[list[_TabledataDataRowColumnDict]]
    """Columns"""

class _TabledataDataCrossSectionalMultiDimensionalCubeRowDict(TypedDict):
    """Type definition for \
        _TabledataDataCrossSectionalMultiDimensionalCubeDict
    """
    rowNo: int
    """Row number"""
    rowText: str
    """Row text"""
    uoM: str
    """Unit of measurement"""
    footnote: str
    """Footnote"""
    columns: NotRequired[
        list[_TabledataDataRowColumnDict | _TabledataDataRowColumnColumnDict]
    ]
    """Columns"""

class _TabledataDataTimeseriesDict(TypedDict):
    """Type definition for _TabledataDict

    All fields are required. But they are typed as `NotRequired` because
    when the endpoint returns empty data, then this dict has no keys at all.
    """
    id: NotRequired[str]
    """ID"""
    title: NotRequired[str]
    """Title"""
    footnote: NotRequired[str]
    """Footnote"""
    frequency: NotRequired[str]
    """Frequency"""
    datasource: NotRequired[str]
    """Data source"""
    generatedBy: NotRequired[str]
    """Generated by"""
    dataLastUpdated: NotRequired[date]
    """Data last updated"""
    dateGenerated: NotRequired[date]
    """Date generated"""
    offset: NotRequired[int | None]
    """Offset"""
    limit: NotRequired[int]
    """Limit"""
    sortBy: NotRequired[str | None]
    """Sort by"""
    timeFilter: NotRequired[str | None]
    """Time filter"""
    between: NotRequired[tuple[int, ...] | None]
    """Between"""
    search: NotRequired[str | None]
    """Search"""
    row: NotRequired[list[_TabledataDataTimeseriesRowDict]]
    """Row"""

class _TabledataDataCrossSectionalMultiDimensionalCubeDict(TypedDict):
    """Type definition for TabledataDict

    All fields are required. But they are typed as `NotRequired` because
    when the endpoint returns empty data, then this dict has no keys at all.
    """
    id: NotRequired[str]
    """ID"""
    tableType: NotRequired[str]
    """Table type"""
    title: NotRequired[str]
    """Title"""
    footnote: NotRequired[str]
    """Footnote"""
    frequency: NotRequired[str]
    """Frequency"""
    dataSource: NotRequired[str]
    """Data source"""
    generatedBy: NotRequired[str]
    """Generated by"""
    dataLastUpdated: NotRequired[date]
    """Data last updated"""
    dateGenerated: NotRequired[date]
    """Date generated"""
    offset: NotRequired[int | None]
    """Offset"""
    limit: NotRequired[int]
    """Limit"""
    between: NotRequired[tuple[int, ...] | None]
    """Between"""
    search: NotRequired[str | None]
    """Search"""
    row: NotRequired[list[_TabledataDataCrossSectionalMultiDimensionalCubeRowDict]]
    """Row"""

class TabledataDict(TypedDict):
    """Type definition for tabledata()"""
    Data: _TabledataDataTimeseriesDict | _TabledataDataCrossSectionalMultiDimensionalCubeDict
    """Data"""
    DataCount: int
    """Data count"""
    StatusCode: int
    """Status code"""
    Message: str
    """Message"""

__all__ = [
    'Url',
    'MetadataDict',
    'ResourceIdDict',
    'TabledataDict',
]
